# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from Anthropic."""

import datetime
import functools
import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import anthropic as anthropic_conversion  # pylint: disable=unused-import
from langfun.core.llms import rest
import pyglove as pg


class AnthropicModelInfo(lf.ModelInfo):
  """Anthropic model info."""

  # Constants for supported MIME types.
  INPUT_IMAGE_TYPES = [
      'image/png',
      'image/jpeg',
      'image/gif',
      'image/webp',
  ]
  INPUT_DOC_TYPES = [
      'application/pdf',
  ]

  LINKS = dict(
      models='https://docs.anthropic.com/claude/docs/models-overview',
      pricing='https://www.anthropic.com/pricing#anthropic-api',
      rate_limits='https://docs.anthropic.com/en/api/rate-limits',
      error_codes='https://docs.anthropic.com/en/api/errors',
  )

  class RateLimits(lf.ModelInfo.RateLimits):
    """Rate limits for Anthropic models."""

    max_input_tokens_per_minute: int
    max_output_tokens_per_minute: int

    @property
    def max_tokens_per_minute(self) -> int:
      return (self.max_input_tokens_per_minute
              + self.max_output_tokens_per_minute)


SUPPORTED_MODELS = [
    AnthropicModelInfo(
        model_id='claude-4-opus-20250514',
        provider='Anthropic',
        in_service=True,
        description='Claude 4 Opus model (5/14/2025).',
        release_date=datetime.datetime(2025, 5, 14),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.5,
            cost_per_1m_input_tokens=15,
            cost_per_1m_output_tokens=75,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_input_tokens_per_minute=100_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-4-sonnet-20250514',
        provider='Anthropic',
        in_service=True,
        description='Claude 4 Sonnet model (5/14/2025).',
        release_date=datetime.datetime(2025, 5, 14),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_input_tokens_per_minute=100_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    # 3.5 Sonnet models.
    AnthropicModelInfo(
        model_id='claude-3-5-sonnet-latest',
        alias_for='claude-3-5-sonnet-20241022',
        provider='Anthropic',
        in_service=True,
        description='Claude 3.5 Sonnet model (latest).',
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-7-sonnet-20250219',
        provider='Anthropic',
        in_service=True,
        description='Claude 3.7 Sonnet model (2/19/2025).',
        release_date=datetime.datetime(2025, 2, 19),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=2000,
            max_input_tokens_per_minute=100_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-5-sonnet-20241022',
        provider='Anthropic',
        in_service=True,
        description='Claude 3.5 Sonnet model (10/22/2024).',
        release_date=datetime.datetime(2024, 10, 22),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-opus-4@20250514',
        alias_for='claude-opus-4-20250514',
        provider='VertexAI',
        in_service=True,
        description='Claude 4 Opus model served on VertexAI (5/14/2025).',
        release_date=datetime.datetime(2025, 5, 14),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.5,
            cost_per_1m_input_tokens=15.0,
            cost_per_1m_output_tokens=75.0,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            max_requests_per_minute=100,
            max_input_tokens_per_minute=1_000_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-sonnet-4@20250514',
        alias_for='claude-sonnet-4@20250514',
        provider='VertexAI',
        in_service=True,
        description='Claude 4 Sonnet model served on VertexAI (5/14/2025).',
        release_date=datetime.datetime(2025, 5, 14),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3.0,
            cost_per_1m_output_tokens=15.0,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            max_requests_per_minute=100,
            max_input_tokens_per_minute=1_000_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-5-sonnet-v2@20241022',
        alias_for='claude-3-5-sonnet-20241022',
        provider='VertexAI',
        in_service=True,
        description='Claude 3.5 Sonnet model served on VertexAI (10/22/2024).',
        release_date=datetime.datetime(2024, 10, 22),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            max_requests_per_minute=100,
            max_input_tokens_per_minute=1_000_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-7-sonnet@20250219',
        alias_for='claude-3-7-sonnet-20250219',
        provider='VertexAI',
        in_service=True,
        description='Claude 3.7 Sonnet model served on VertexAI (02/19/2025).',
        release_date=datetime.datetime(2025, 2, 19),
        input_modalities=(
            AnthropicModelInfo.INPUT_IMAGE_TYPES
            + AnthropicModelInfo.INPUT_DOC_TYPES
        ),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.3,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            max_requests_per_minute=100,
            max_input_tokens_per_minute=1_000_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    # 3.5 Haiku models.
    AnthropicModelInfo(
        model_id='claude-3-5-haiku-latest',
        alias_for='claude-3-5-haiku-20241022',
        provider='Anthropic',
        in_service=True,
        description='Claude 3.5 Haiku v2 model (10/22/2024).',
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.08,
            cost_per_1m_input_tokens=0.8,
            cost_per_1m_output_tokens=4,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-5-haiku-20241022',
        provider='Anthropic',
        in_service=True,
        description='Claude 3.5 Haiku v2 model (10/22/2024).',
        release_date=datetime.datetime(2024, 10, 22),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.08,
            cost_per_1m_input_tokens=0.8,
            cost_per_1m_output_tokens=4,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-5-haiku@20241022',
        alias_for='claude-3-5-haiku-20241022',
        provider='VertexAI',
        in_service=True,
        description='Claude 3.5 Haiku model served on VertexAI (10/22/2024).',
        release_date=datetime.datetime(2024, 10, 22),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.08,
            cost_per_1m_input_tokens=0.8,
            cost_per_1m_output_tokens=4,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    # 3.0 Opus models.
    AnthropicModelInfo(
        model_id='claude-3-opus-latest',
        alias_for='claude-3-opus-20240229',
        provider='Anthropic',
        in_service=True,
        description='Claude 3 Opus model (latest).',
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.5,
            cost_per_1m_input_tokens=15,
            cost_per_1m_output_tokens=75,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-opus-20240229',
        provider='Anthropic',
        in_service=True,
        description='Claude 3 Opus model (02/29/2024).',
        release_date=datetime.datetime(2024, 2, 29),
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.5,
            cost_per_1m_input_tokens=15,
            cost_per_1m_output_tokens=75,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-opus@20240229',
        alias_for='claude-3-opus-20240229',
        provider='VertexAI',
        in_service=True,
        description='Claude 3 Opus model served on VertexAI (02/29/2024).',
        release_date=datetime.datetime(2024, 2, 29),
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=1.5,
            cost_per_1m_input_tokens=15,
            cost_per_1m_output_tokens=75,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    # 3.0 Sonnet models.
    AnthropicModelInfo(
        model_id='claude-3-sonnet-20240229',
        provider='Anthropic',
        in_service=True,
        description='Claude 3 Sonnet model (02/29/2024).',
        release_date=datetime.datetime(2024, 2, 29),
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-sonnet@20240229',
        alias_for='claude-3-sonnet-20240229',
        provider='VertexAI',
        in_service=True,
        description='Claude 3 Sonnet model served on VertexAI (02/29/2024).',
        release_date=datetime.datetime(2024, 2, 29),
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=3,
            cost_per_1m_output_tokens=15,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    # 3.0 Haiku models.
    AnthropicModelInfo(
        model_id='claude-3-haiku-20240307',
        provider='Anthropic',
        in_service=True,
        description='Claude 3 Haiku model (03/07/2024).',
        release_date=datetime.datetime(2024, 3, 7),
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=0.25,
            cost_per_1m_output_tokens=1.25,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
    AnthropicModelInfo(
        model_id='claude-3-haiku@20240307',
        alias_for='claude-3-haiku-20240307',
        provider='Anthropic',
        in_service=True,
        description='Claude 3 Haiku model served on VertexAI (03/07/2024).',
        release_date=datetime.datetime(2024, 3, 7),
        input_modalities=AnthropicModelInfo.INPUT_IMAGE_TYPES,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=200_000,
            max_output_tokens=4_096,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=None,
            cost_per_1m_input_tokens=0.25,
            cost_per_1m_output_tokens=1.25,
        ),
        rate_limits=AnthropicModelInfo.RateLimits(
            # Tier 4 rate limits
            max_requests_per_minute=4000,
            max_input_tokens_per_minute=400_000,
            max_output_tokens_per_minute=80_000,
        ),
    ),
]


_SUPPORTED_MODELS_BY_MODEL_ID = {m.model_id: m for m in SUPPORTED_MODELS}


@lf.use_init_args(['model'])
class Anthropic(rest.REST):
  """Anthropic LLMs (Claude) through REST APIs.

  See https://docs.anthropic.com/claude/reference/messages_post
  """

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, [m.model_id for m in SUPPORTED_MODELS]
      ),
      'The name of the model to use.',
  ]

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'ANTHROPIC_API_KEY'."
      ),
  ] = None

  api_endpoint: str = 'https://api.anthropic.com/v1/messages'

  api_version: Annotated[
      str,
      'Anthropic API version.'
  ] = '2023-06-01'

  def _on_bound(self):
    super()._on_bound()
    self._api_key = None

  def _initialize(self):
    api_key = self.api_key or os.environ.get('ANTHROPIC_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `ANTHROPIC_API_KEY` with your Anthropic API key.'
      )
    self._api_key = api_key

  @property
  def headers(self) -> dict[str, Any]:
    return {
        'x-api-key': self._api_key,
        'anthropic-version': self.api_version,
        'content-type': 'application/json',
        'anthropic-beta': 'output-128k-2025-02-19',
    }

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    mi = _SUPPORTED_MODELS_BY_MODEL_ID[self.model]
    if mi.provider != 'Anthropic':
      assert mi.alias_for is not None
      mi = _SUPPORTED_MODELS_BY_MODEL_ID[mi.alias_for]
      assert mi.provider == 'Anthropic', mi
    return mi

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request = dict()
    request.update(self._request_args(sampling_options))

    def modality_check(chunk: Any) -> Any:
      if isinstance(chunk, lf_modalities.Mime):
        if not self.supports_input(chunk.mime_type):
          raise ValueError(f'Unsupported modality: {chunk!r}.')
      return chunk

    if system_message := prompt.get('system_message'):
      assert isinstance(system_message, lf.SystemMessage), type(system_message)
      request['system'] = system_message.text

    messages = [
        prompt.as_format('anthropic', chunk_preprocessor=modality_check)
    ]
    request.update(messages=messages)
    return request

  def _request_args(self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # Authropic requires `max_tokens` to be specified.
    max_tokens = (
        options.max_tokens or self.model_info.context_length.max_output_tokens
    )
    args = dict(
        model=self.model,
        max_tokens=max_tokens,
        stream=False,
    )
    if options.stop:
      args['stop_sequences'] = options.stop
    if options.temperature is not None:
      args['temperature'] = options.temperature
    if options.top_k is not None:
      args['top_k'] = options.top_k
    if options.top_p is not None:
      args['top_p'] = options.top_p
    if options.max_thinking_tokens is not None:
      args['thinking'] = {
          'type': 'enabled',
          # Minimum budget is 1,024 tokens.
          'budget_tokens': options.max_thinking_tokens,
      }
      # max_tokens, which is thinking tokens + response tokens, must be greater
      # than the thinking tokens.
      if args['max_tokens'] < options.max_thinking_tokens:
        args['max_tokens'] += options.max_thinking_tokens

      # Thinking isn’t compatible with temperature, top_p, or top_k.
      # https://docs.anthropic.com/en/docs/build-with-claude/extended-thinking#important-considerations-when-using-extended-thinking
      args.pop('temperature', None)
      args.pop('top_k', None)
      args.pop('top_p', None)
    return args

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    message = lf.Message.from_value(json, format='anthropic')
    input_tokens = json['usage']['input_tokens']
    output_tokens = json['usage']['output_tokens']
    return lf.LMSamplingResult(
        [lf.LMSample(message)],
        usage=lf.LMSamplingUsage(
            prompt_tokens=input_tokens,
            completion_tokens=output_tokens,
            total_tokens=input_tokens + output_tokens,
        ),
    )

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if status_code == 413 and b'Prompt is too long' in content:
      return lf.ContextLimitError(f'{status_code}: {content}')
    return super()._error(status_code, content)


class Claude4(Anthropic):
  """Base class for Claude 4 models."""


# pylint: disable=invalid-name
class Claude4Opus_20250514(Claude4):
  """Claude 4 Opus model 20250514."""

  model = 'claude-4-opus-20250514'


# pylint: disable=invalid-name
class Claude4Sonnet_20250514(Claude4):
  """Claude 4 Sonnet model 20250514."""

  model = 'claude-4-sonnet-20250514'


class Claude37(Anthropic):
  """Base class for Claude 3.7 models."""


# pylint: disable=invalid-name
class Claude37Sonnet_20250219(Claude37):
  """Claude 3.7 Sonnet model (latest)."""

  model = 'claude-3-7-sonnet-20250219'


class Claude35(Anthropic):
  """Base class for Claude 3.5 models."""


class Claude35Sonnet(Claude35):
  """Claude 3.5 Sonnet model (latest)."""
  model = 'claude-3-5-sonnet-latest'


class Claude35Sonnet_20241022(Claude35):  # pylint: disable=invalid-name
  """Claude 3.5 Sonnet model (10/22/2024)."""
  model = 'claude-3-5-sonnet-20241022'


class Claude35Haiku(Claude35):
  """Claude 3.5 Haiku model (latest)."""
  model = 'claude-3-5-haiku-latest'


class Claude35Haiku_20241022(Claude35):  # pylint: disable=invalid-name
  """Claude 3.5 Haiku model (10/22/2024)."""
  model = 'claude-3-5-haiku-20241022'


class Claude3(Anthropic):
  """Base class for Claude 3 models."""


class Claude3Opus(Claude3):
  """Claude 3 Opus model (latest)."""

  model = 'claude-3-opus-latest'


class Claude3Opus_20240229(Claude3):  # pylint: disable=invalid-name
  """Claude 3 Opus model (02/29/2024)."""

  model = 'claude-3-opus-20240229'


class Claude3Sonnet(Claude3):
  """Claude 3 Sonnet model."""

  model = 'claude-3-sonnet-20240229'


class Claude3Sonnet_20240229(Claude3):  # pylint: disable=invalid-name
  """Claude 3 Sonnet model (02/29/2024)."""

  model = 'claude-3-sonnet-20240229'


class Claude3Haiku(Claude3):
  """Claude 3 Haiku model."""

  model = 'claude-3-haiku-20240307'


class Claude3Haiku_20240307(Claude3):  # pylint: disable=invalid-name
  """Claude 3 Haiku model (03/07/2024)."""

  model = 'claude-3-haiku-20240307'


def _register_anthropic_models():
  """Registers Anthropic models."""
  for m in SUPPORTED_MODELS:
    if m.provider == 'Anthropic':
      lf.LanguageModel.register(m.model_id, Anthropic)

_register_anthropic_models()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OpenAI models hosted on Azure."""
import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core.llms import openai
import pyglove as pg


@lf.use_init_args(['model', 'deployment_name'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class AzureOpenAI(openai.OpenAI):
  """Azure OpenAI model service.

  This service interacts with the Azure OpenAI API to generate chat completions.
  It uses the deployment_name and API version to construct the endpoint, and
  authenticates using an API key provided via parameter or the
  AZURE_OPENAI_API_KEY environment variable.

  Example:
      lm = AzureOpenAI(
          model='gpt-4o',
          deployment_name='gpt-4o',
          api_version='2024-08-01-preview',
          azure_endpoint='https://trackname.openai.azure.com/',
          api_key='token'
      )
      response = lf.query(prompt="what the capital of France", lm=lm)
      print(response)
  """

  deployment_name: Annotated[
      str,
      'The name of the Azure OpenAI deployment.'
  ] = 'gpt-4'

  api_version: Annotated[
      str,
      'The API version for Azure OpenAI.'
  ] = '2023-05-15'

  azure_endpoint: Annotated[
      str,
      (
          'The base URL for Azure OpenAI '
          '(e.g. "https://<your-resource>.openai.azure.com/")'
      )
  ] = 'https://api.openai.azure.com/'

  def _initialize(self):
    # Authentication
    self._api_key = self.api_key or os.environ.get('AZURE_OPENAI_API_KEY')
    if not self._api_key:
      raise ValueError(
          'Azure OpenAI requires an API key. Please provide '
          'via `api_key` argument or AZURE_OPENAI_API_KEY '
          'environment variable.'
      )

    # Endpoint construction
    self._api_endpoint = (
        f"{self.azure_endpoint.rstrip('/')}/openai/deployments/"
        f"{self.deployment_name}/chat/completions"
        f"?api-version={self.api_version}"
    )

  @property
  def api_endpoint(self) -> str:
    return self._api_endpoint

  @property
  def headers(self) -> dict[str, Any]:
    headers = super().headers
    headers.update({
        'api-key': self._api_key,
    })
    return headers



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from Groq."""

import datetime
import functools
import os
from typing import Annotated, Any, Final

import langfun.core as lf
from langfun.core.llms import openai_compatible
import pyglove as pg


class GroqModelInfo(lf.ModelInfo):
  """Groq model info."""

  LINKS = dict(
      models='https://console.groq.com/docs/models',
      pricing='https://groq.com/pricing/',
      rate_limits='https://console.groq.com/docs/rate-limits',
      error_codes='https://console.groq.com/docs/errors',
  )

  provider: Final[str] = 'Groq'  # pylint: disable=invalid-name


SUPPORTED_MODELS = [
    #
    # Llama models.
    #
    GroqModelInfo(
        model_id='llama-3.3-70b-versatile',
        in_service=True,
        model_type='instruction-tuned',
        description='Llama 3.3 70B model on Groq (Production)',
        url='https://huggingface.co/meta-llama/Llama-3.3-70B-Instruct',
        release_date=datetime.datetime(2024, 12, 6),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.59,
            cost_per_1m_output_tokens=0.79,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=1_000,
            max_tokens_per_minute=120_000,
        ),
    ),
    GroqModelInfo(
        model_id='llama-3.3-70b-specdec',
        in_service=True,
        model_type='instruction-tuned',
        description='Llama 3.3 70B model on Groq (Production)',
        url='https://huggingface.co/meta-llama/Llama-3.3-70B-Instruct',
        release_date=datetime.datetime(2024, 12, 6),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=8_192,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.59,
            cost_per_1m_output_tokens=0.99,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=30_000,
        ),
    ),
    GroqModelInfo(
        model_id='llama-3.2-1b-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='Llama 3.2 1B model on Groq (Preview)',
        url='https://huggingface.co/meta-llama/Llama-3.2-1B',
        release_date=datetime.datetime(2024, 12, 6),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.04,
            cost_per_1m_output_tokens=0.04,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=30_000,
        ),
    ),
    GroqModelInfo(
        model_id='llama-3.2-3b-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='Llama 3.2 3B model on Groq (Preview)',
        url='https://huggingface.co/meta-llama/Llama-3.2-3B',
        release_date=datetime.datetime(2024, 12, 6),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.06,
            cost_per_1m_output_tokens=0.06,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=30_000,
        ),
    ),
    GroqModelInfo(
        model_id='llama-3.2-11b-vision-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='Llama 3.2 11B vision model on Groq (Preview)',
        url='https://huggingface.co/meta-llama/Llama-3.2-11B-Vision',
        release_date=datetime.datetime(2024, 12, 6),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.18,
            cost_per_1m_output_tokens=0.18,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=30_000,
        ),
    ),
    GroqModelInfo(
        model_id='llama-3.2-90b-vision-preview',
        in_service=True,
        model_type='instruction-tuned',
        description='Llama 3.2 90B vision model on Groq (Preview)',
        url='https://huggingface.co/meta-llama/Llama-3.2-90B-Vision',
        release_date=datetime.datetime(2024, 12, 6),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.9,
            cost_per_1m_output_tokens=0.9,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=30_000,
        ),
    ),
    #
    # DeepSeek models
    #
    GroqModelInfo(
        model_id='deepseek-r1-distill-llama-70b',
        in_service=True,
        model_type='thinking',
        description='DeepSeek R1 distilled from Llama 70B (Preview)',
        url='https://console.groq.com/docs/model/deepseek-r1-distill-llama-70b',
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        # TODO(daiyip): Pricing needs to be computed based on the number of
        # input/output tokens.
        pricing=None,
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=1_000,
            max_tokens_per_minute=120_000,
        ),
    ),
    GroqModelInfo(
        model_id='deepseek-r1-distill-llama-70b-specdec',
        in_service=True,
        model_type='thinking',
        description='DeepSeek R1 distilled from Llama 70B (Preview)',
        url='https://console.groq.com/docs/model/deepseek-r1-distill-llama-70b',
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=16_384,
        ),
        # TODO(daiyip): Pricing needs to be computed based on the number of
        # input/output tokens.
        pricing=None,
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=60_000,
        ),
    ),
    #
    # Gemma models.
    #
    GroqModelInfo(
        model_id='gemma2-9b-it',
        in_service=True,
        model_type='instruction-tuned',
        description='Google Gemma 2 9B model on Groq.',
        url='https://huggingface.co/google/gemma-2-9b-it',
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=8_192,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.2,
            cost_per_1m_output_tokens=0.2,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=200,
            max_tokens_per_minute=30_000,
        ),
    ),
    #
    # Mixtral models.
    #
    GroqModelInfo(
        model_id='mixtral-8x7b-32768',
        in_service=True,
        model_type='instruction-tuned',
        description='Mixtral 8x7B model on Groq (Production)',
        url='https://huggingface.co/mistralai/Mixtral-8x7B-Instruct-v0.1',
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=32_768,
            max_output_tokens=None,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.24,
            cost_per_1m_output_tokens=0.24,
        ),
        rate_limits=lf.ModelInfo.RateLimits(
            # Developer tier.
            max_requests_per_minute=100,
            max_tokens_per_minute=25_000,
        ),
    ),
]

_SUPPORTED_MODELS_BY_ID = {m.model_id: m for m in SUPPORTED_MODELS}


@lf.use_init_args(['model'])
class Groq(openai_compatible.OpenAICompatible):
  """Groq LLMs through REST APIs (OpenAI compatible).

  See https://platform.openai.com/docs/api-reference/chat
  """

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, [m.model_id for m in SUPPORTED_MODELS]
      ),
      'The name of the model to use.',
  ]

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'GROQ_API_KEY'."
      ),
  ] = None

  api_endpoint: str = 'https://api.groq.com/openai/v1/chat/completions'

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    return _SUPPORTED_MODELS_BY_ID[self.model]

  @property
  def headers(self) -> dict[str, Any]:
    api_key = self.api_key or os.environ.get('GROQ_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `GROQ_API_KEY` with your Groq API key.'
      )
    headers = super().headers
    headers.update({
        'Authorization': f'Bearer {api_key}',
    })
    return headers

  def _request_args(self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # `logprobs` and `top_logprobs` flags are not supported on Groq yet.
    args = super()._request_args(options)
    args.pop('logprobs', None)
    args.pop('top_logprobs', None)
    return args


class GroqLlama33_70B_Versatile(Groq):  # pylint: disable=invalid-name
  """Llama3.2-3B with 128K context window."""
  model = 'llama-3.3-70b-versatile'


class GroqLlama33_70B_SpecDec(Groq):  # pylint: disable=invalid-name
  """Llama3.3-70B with 8K context window."""
  model = 'llama-3.3-70b-specdec'


class GroqLlama32_1B(Groq):  # pylint: disable=invalid-name
  """Llama3.2-1B."""
  model = 'llama-3.2-1b-preview'


class GroqLlama32_3B(Groq):  # pylint: disable=invalid-name
  """Llama3.2-3B."""
  model = 'llama-3.2-3b-preview'


class GroqLlama32_11B_Vision(Groq):  # pylint: disable=invalid-name
  """Llama3.2-11B vision."""
  model = 'llama-3.2-11b-vision-preview'


class GroqLlama32_90B_Vision(Groq):  # pylint: disable=invalid-name
  """Llama3.2-90B vision."""
  model = 'llama-3.2-90b-vision-preview'


class GroqDeepSeekR1_DistillLlama_70B(Groq):  # pylint: disable=invalid-name
  """DeepSeek R1 distilled from Llama 70B."""
  model = 'deepseek-r1-distill-llama-70b'


class GroqDeepSeekR1_DistillLlama_70B_SpecDec(Groq):  # pylint: disable=invalid-name
  """DeepSeek R1 distilled from Llama 70B (SpecDec)."""
  model = 'deepseek-r1-distill-llama-70b-specdec'


class GroqMistral_8x7B(Groq):  # pylint: disable=invalid-name
  """Mixtral 8x7B."""
  model = 'mixtral-8x7b-32768'


class GroqGemma2_9B_IT(Groq):  # pylint: disable=invalid-name
  """Gemma2 9B."""
  model = 'gemma2-9b-it'


#
# Register Groq models so they can be retrieved with LanguageModel.get().
#


def _groq_model(model: str, *args, **kwargs):
  model = model.removeprefix('groq://')
  return Groq(model, *args, **kwargs)


def _register_groq_models():
  """Registers Groq models."""
  for m in SUPPORTED_MODELS:
    lf.LanguageModel.register('groq://' + m.model_id, _groq_model)

_register_groq_models()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test sampling."""

import unittest

from langfun.core import component
from langfun.core import language_model
from langfun.core import sampling
from langfun.core.langfunc import LangFunc
import pyglove as pg


class ExcitedEchoer(language_model.LanguageModel):
  """LM for testing."""

  def _sample(self, prompts) -> list[language_model.LMSamplingResult]:
    return [
        language_model.LMSamplingResult([
            language_model.LMSample(prompt.text + '!!!')
            ]) for prompt in prompts
    ]


class SamplingTest(unittest.TestCase):

  def test_sweep(self):
    l = LangFunc('Compute {{x}} and {{y}}', x=pg.oneof([1, 2]))
    with component.context(lm=ExcitedEchoer()):
      samples = list(sampling.sweep(l, y=pg.oneof([3, 4])))
      samples = sorted(samples, key=lambda x: (x[0].x, x[0].y))

    self.assertEqual(
        samples,
        [
            ('Compute 1 and 3', 'Compute 1 and 3!!!'),
            ('Compute 1 and 4', 'Compute 1 and 4!!!'),
            ('Compute 2 and 3', 'Compute 2 and 3!!!'),
            ('Compute 2 and 4', 'Compute 2 and 4!!!'),
        ],
    )

  def test_random_sample(self):
    l = LangFunc('Compute {{x}} and {{y}}', x=pg.oneof([1, 3, 5]))
    with component.context(lm=ExcitedEchoer()):
      samples = list(
          sampling.random_sample(l, y=pg.oneof([2, 4]), num_examples=3, seed=1)
      )
      samples = sorted(samples, key=lambda x: (x[0].x, x[0].y))

    self.assertEqual(
        samples,
        [
            ('Compute 1 and 2', 'Compute 1 and 2!!!'),
            ('Compute 3 and 2', 'Compute 3 and 2!!!'),
            ('Compute 3 and 4', 'Compute 3 and 4!!!'),
        ],
    )

  def test_random_sample_with_const_space(self):
    l = LangFunc('Compute {{x}} and {{y}}', x=1)

    with component.context(lm=ExcitedEchoer()):
      samples = list(sampling.random_sample(l, y=2, num_examples=3, seed=1))
      self.assertEqual(
          samples,
          [
              ('Compute 1 and 2', 'Compute 1 and 2!!!'),
              ('Compute 1 and 2', 'Compute 1 and 2!!!'),
              ('Compute 1 and 2', 'Compute 1 and 2!!!'),
          ],
      )

  def test_random_sample_with_errors(self):
    l = LangFunc('Compute {{x}} and {{y}}', x=pg.oneof([1, 3, 5]))
    with self.assertRaisesRegex(AttributeError, '`lm` is not found'):
      _ = next(sampling.random_sample(l, y=pg.oneof([2, 4]), num_examples=1))

  def test_random_sample_with_silenced_errors(self):
    l = LangFunc('Compute {{x}} and {{y}}', x=pg.oneof([1, 3, 5]))
    samples = list(sampling.random_sample(
        l,
        y=pg.oneof([2, 4]),
        num_examples=3,
        seed=1,
        silence_on_errors=(AttributeError,),
        ignore_examples_with_errors=False))

    samples = sorted(samples, key=lambda x: (x[0].x, x[0].y))
    self.assertEqual(
        [x[0] for x in samples],
        [
            'Compute 1 and 2',
            'Compute 3 and 2',
            'Compute 3 and 4',
        ]
    )
    for x in samples:
      self.assertIsInstance(x[1], AttributeError)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for agentic actions."""

import abc
import contextlib
import functools
import threading
import time
import typing
from typing import Annotated, Any, Callable, Iterable, Iterator, Optional, Type, Union
import uuid
import langfun.core as lf
from langfun.core import structured as lf_structured
import pyglove as pg


class ActionError(Exception):   # pylint: disable=g-bad-exception-name
  """Base class for common action errors."""


class ActionTimeoutError(ActionError):
  """Raised when an action exceeds the max execution time."""


class Action(pg.Object):
  """Base class for Langfun's agentic actions.

  # Developing Actions

  In Langfun, an `Action` is a class representing a task an agent can execute.
  To define custom actions, subclass `lf.agentic.Action` and implement the
  `call` method, which contains the logic for the action's execution.

  ```python
  class Calculate(lf.agentic.Action):
    expression: str

    def call(self, session: Session, *, lm: lf.LanguageModel, **kwargs):
      return session.query(expression, float, lm=lm)
  ```

  Key aspects of the `call` method:

  - `session` (First Argument): An `lf.Session` object required to make queries,
    perform logging, and add metadata to the action. It also tracks the
    execution of the action and its sub-actions.

    - Use `session.query(...)` to make calls to a Language Model.
    - Use `session.debug(...)`, `session.info(...)`, `session.warning(...)`,
      and `session.error(...)` for adding logs associated with the
      current action.
    - Use `session.add_metadata(...)` to associate custom metadata with
      the current action.

  - Keyword Arguments (e.g., lm): Arguments required for the action's execution
    (like a language model) should be defined as keyword arguments.

  - **kwargs: Include **kwargs to allow:

    - Users to pass additional arguments to child actions.
    - The action to gracefully handle extra arguments passed by parent actions.

  # Using Actions

  ## Creating Action objects
  Action objects can be instantiated in two primary ways:

  - Direct instantiation by Users:

    ```
    calculate_action = Calculate(expression='1 + 1')
    ```

  - Generation by Language Models (LLMs): LLMs can generate Action objects when
    provided with an "action space" (a schema defining possible actions). The
    LLM populates the action's attributes. User code can then invoke the
    generated action.

  ```python
  import pyglove as pg
  import langfun as lf

  # Define possible actions for the LLM
  class Search(lf.agentic.Action):
    query: str
    def call(self, session: lf.Session, *, lm: lf.LanguageModel, **kwargs):
      # Placeholder for actual search logic
      return f"Results for: {self.query}"

  class DirectAnswer(lf.agentic.Action):
    answer: str
    def call(self, session: lf.Session, *, lm: lf.LanguageModel, **kwargs):
      return self.answer

  # Define the schema for the LLM's output
  class NextStep(pg.Object):
    step_by_step_thoughts: list[str]
    next_action: Calculate | Search | DirectAnswer

  # Query the LLM to determine the next step
  next_step = lf.query(
      'What is the next step for {{question}}?',
      NextStep,
      question='why is the sky blue?'
  )
  # Execute the action chosen by the LLM
  result = next_step.next_action()
  print(result)
  ```

  ## Invoking Actions and Managing Sessions:

  When an action is called, the session argument (the first argument to call)
  is handled as follows:

  - Implicit Session Management: If no session is explicitly provided when
    calling an action, Langfun automatically creates and passes one.

    ```python
    calc = Calculate(expression='1 + 1')

    # A session is implicitly created and passed here.
    result = calc()
    print(result)

    # Access the implicitly created session.
    # print(calc.session)
    ```

  - Explicit Session Management: You can create and manage `lf.Session` objects
    explicitly. This is useful for customizing session identifiers or managing
    a shared context for multiple actions.

  ```python
  calc = Calculate(expression='1 + 1')

  # Explicitly create and pass a session.
  with lf.Session(id='my_agent_session') as session:
    result = calc(session=session) # Pass the session explicitly
    print(result)
  ```

  ## Accessing Execution Trajectory:

  After an action is executed, the Session object holds a record of its
  execution, known as the trajectory. This includes queries made and any
  sub-actions performed.

  - To access all queries issued directly by the root action:

    ```python
    print(session.root.execution.queries)
    ```
  - To access all actions issued by the root action and any of its
    sub-actions (recursively):

    ```python
    print(session.root.execution.all_queries)
    ```
  - To access all child actions issued by the root action:

    ```python
    print(session.root.execution.actions)
    ```

  - To access all the actions in the sub-tree issued by the root action:

    ```python
    print(session.root.execution.all_actions)
    ```
  """

  def _on_bound(self):
    super()._on_bound()
    self._session = None
    self._invocation: ActionInvocation | None = None

  @property
  def session(self) -> Optional['Session']:
    """Returns the session started by this action."""
    return self._session

  @property
  def result(self) -> Any:
    """Returns the result of the action."""
    return self._invocation.result if self._invocation else None

  @property
  def metadata(self) -> dict[str, Any] | None:
    """Returns the metadata associated with the result from previous call."""
    return self._invocation.metadata if self._invocation else None

  @property
  def invocation(self) -> Optional['ActionInvocation']:
    """Returns last invocation. None if the action is not executed."""
    return self._invocation

  async def acall(
      self,
      session: Optional['Session'] = None,
      *,
      max_execution_time: float | None = None,
      show_progress: bool = True,
      verbose: bool = False,
      **kwargs
  ) -> Any:
    """Async version of `__call__`."""
    # TODO(daiyip): implement native async calling.
    return await lf.invoke_async(
        self.__call__,
        session,
        max_execution_time=max_execution_time,
        show_progress=show_progress,
        verbose=verbose,
        **kwargs
    )

  def __call__(
      self,
      session: Optional['Session'] = None,
      *,
      max_execution_time: float | None = None,
      show_progress: bool = True,
      verbose: bool = False,
      **kwargs
  ) -> Any:
    """Executes the action.

    Args:
      session: The session to use for the action.
      max_execution_time: The max allowed execution time in seconds for the
        action. The effective `max_execution_time` is the smaller of the
        remaining time of the parent invocation and the `max_execution_time`
        for the current invocation. Since we are running action within the
        current thread, there is no guarantee that the action will be executed
        within this time limit, but we will try to stop the action as soon as
        the next operation on the session is called.
      show_progress: Whether to show the progress of the action.
      verbose: Whether to log the action execution.
      **kwargs: Additional keyword arguments to pass to the action.

    Returns:
      The result of the action.
    """
    if session is None:
      session = Session(verbose=verbose)
      session.start()

      if show_progress:
        lf.console.display(pg.view(session, name='agent_session'))

      # For the top-level action, we store the session in the metadata.
      self._session = session
    else:
      self._session = None

    with session.track_action(self, max_execution_time=max_execution_time):
      try:
        result = self.call(session=session, **kwargs)
        self._invocation.end(result)
      except BaseException as e:
        error = pg.ErrorInfo.from_exception(e)
        self._invocation.end(result=None, error=error)
        if self._session is not None:
          self._session.end(result=None, error=error)
        raise

    if self._session is not None:
      # Session is created by current action. Stop the session.
      self._session.end(result)
    return result

  @abc.abstractmethod
  def call(self, session: 'Session', **kwargs) -> Any:
    """Calls the action.

    Args:
      session: The session to use for the action.
      **kwargs: Additional keyword arguments to pass to the action.

    Returns:
      The result of the action.
    """


# Type definition for traced item during execution.
TracedItem = Union[
    lf_structured.QueryInvocation,
    'ActionInvocation',
    'ExecutionTrace',
    'ParallelExecutions',
    # NOTE(daiyip): Consider remove log entry once we migrate existing agents.
    lf.logging.LogEntry,
]


class ExecutionTrace(pg.Object, pg.views.html.HtmlTreeView.Extension):
  """Trace of the execution of an action."""

  name: Annotated[
      str | None,
      (
          'The name of the execution trace. If None, the trace is unnamed, '
          'which is the case for the top-level trace of an action. An '
          'execution trace could have sub-traces, called phases, which are '
          'created and named by `session.phase()` context manager.'
      )
  ] = None

  start_time: Annotated[
      float | None,
      'The start time of the execution. If None, the execution is not started.'
  ] = None

  end_time: Annotated[
      float | None,
      'The end time of the execution. If None, the execution is not ended.'
  ] = None

  items: Annotated[
      list[TracedItem],
      'All tracked execution items in the sequence.'
  ] = []

  def _on_bound(self):
    super()._on_bound()
    self._usage_summary = lf.UsageSummary()
    for item in self.items:
      if hasattr(item, 'usage_summary'):
        self._usage_summary.merge(item.usage_summary)

    self._tab_control = None
    self._time_badge = None

  def _on_parent_change(self, *args, **kwargs):
    super()._on_parent_change(*args, **kwargs)
    self.__dict__.pop('id', None)

  def indexof(self, item: TracedItem, count_item_cls: Type[Any]) -> int:
    """Returns the index of the child items of given type."""
    pos = 0
    for x in self._iter_children(count_item_cls):
      if x is item:
        return pos
      pos += 1
    return -1

  @functools.cached_property
  def id(self) -> str:
    parent = self.sym_parent
    if isinstance(parent, ActionInvocation):
      # Current execution trace is the body of an action.
      return parent.id
    elif isinstance(parent, pg.List):
      container = parent.sym_parent
      if isinstance(container, ExecutionTrace):
        # Current execution trace is a phase.
        group_id = (
            self.name or f'g{container.indexof(self, ExecutionTrace) + 1}'
        )
        return f'{container.id}/{group_id}'
      elif isinstance(container, ParallelExecutions):
        # Current execution trace is a parallel branch.
        return f'{container.id}/b{self.sym_path.key + 1}'
    return ''

  def reset(self) -> None:
    """Resets the execution trace."""
    self.rebind(items=[], skip_notification=True, raise_on_no_change=False)

  def start(self) -> None:
    assert self.start_time is None, 'Execution already started.'
    self.rebind(start_time=time.time(), skip_notification=True)
    if self._time_badge is not None:
      self._time_badge.update(
          'Starting',
          add_class=['starting'],
          remove_class=['not-started'],
      )

  def stop(self) -> None:
    assert self.end_time is None, 'Execution already stopped.'
    self.rebind(end_time=time.time(), skip_notification=True)
    if self._time_badge is not None:
      self._time_badge.update(
          f'{int(self.elapse)} seconds',
          tooltip=pg.format(self.execution_summary(), verbose=False),
          add_class=['finished'],
          remove_class=['running'],
      )

  @property
  def has_started(self) -> bool:
    return self.start_time is not None

  @property
  def has_stopped(self) -> bool:
    return self.end_time is not None

  @property
  def elapse(self) -> float:
    """Returns the elapsed time of the execution."""
    if self.start_time is None:
      return 0.0
    if self.end_time is None:
      return time.time() - self.start_time
    return self.end_time - self.start_time

  @property
  def queries(self) -> list[lf_structured.QueryInvocation]:
    """Returns queries from the sequence."""
    return list(self._iter_children(lf_structured.QueryInvocation))

  @property
  def actions(self) -> list['ActionInvocation']:
    """Returns action invocations from the sequence."""
    return list(self._iter_children(ActionInvocation))

  @property
  def logs(self) -> list[lf.logging.LogEntry]:
    """Returns logs from the sequence."""
    return list(self._iter_children(lf.logging.LogEntry))

  @property
  def all_queries(self) -> list[lf_structured.QueryInvocation]:
    """Returns all queries from current trace and its child execution items."""
    return list(self._iter_subtree(lf_structured.QueryInvocation))

  @property
  def all_actions(self) -> list['ActionInvocation']:
    """Returns all actions from current trace and its child execution items."""
    return list(self._iter_subtree(ActionInvocation))

  @property
  def all_logs(self) -> list[lf.logging.LogEntry]:
    """Returns all logs from current trace and its child execution items."""
    return list(self._iter_subtree(lf.logging.LogEntry))

  def _iter_children(self, item_cls: Type[Any]) -> Iterator[TracedItem]:
    for item in self.items:
      if isinstance(item, item_cls):
        yield item
      elif isinstance(item, ExecutionTrace):
        for x in item._iter_children(item_cls):  # pylint: disable=protected-access
          yield x
      elif isinstance(item, ParallelExecutions):
        for branch in item.branches:
          for x in branch._iter_children(item_cls):  # pylint: disable=protected-access
            yield x

  def _iter_subtree(self, item_cls: Type[Any]) -> Iterator[TracedItem]:
    for item in self.items:
      if isinstance(item, item_cls):
        yield item
      if isinstance(item, ActionInvocation):
        for x in item.execution._iter_subtree(item_cls):  # pylint: disable=protected-access
          yield x
      elif isinstance(item, ExecutionTrace):
        for x in item._iter_subtree(item_cls):  # pylint: disable=protected-access
          yield x
      elif isinstance(item, ParallelExecutions):
        for branch in item.branches:
          for x in branch._iter_subtree(item_cls):  # pylint: disable=protected-access
            yield x

  #
  # Shortcut methods to operate on the execution trace.
  #

  def __len__(self) -> int:
    return len(self.items)

  def __iter__(self) -> Iterator[TracedItem]:
    return iter(self.items)

  def __bool__(self) -> bool:
    return bool(self.items)

  def __getitem__(self, index: int) -> TracedItem:
    return self.items[index]

  def merge_usage_summary(self, usage_summary: lf.UsageSummary) -> None:
    if usage_summary.total.num_requests == 0:
      return
    current_invocation = self
    while current_invocation is not None:
      current_invocation.usage_summary.merge(usage_summary)
      current_invocation = typing.cast(
          ExecutionTrace,
          current_invocation.sym_ancestor(
              lambda x: isinstance(x, ExecutionTrace)
          )
      )

  def append(self, item: TracedItem) -> None:
    """Appends an item to the sequence."""
    with pg.notify_on_change(False):
      self.items.append(item)

    if isinstance(item, lf_structured.QueryInvocation):
      self.merge_usage_summary(item.usage_summary)

    if self._tab_control is not None:
      self._tab_control.append(self._execution_item_tab(item))

    if (self._time_badge is not None
        and not isinstance(item, lf.logging.LogEntry)):
      sub_task_label = self._execution_item_label(item)
      self._time_badge.update(
          text=sub_task_label.text,
          tooltip=sub_task_label.tooltip.content,
          add_class=['running'],
          remove_class=['not-started'],
      )

  def extend(self, items: Iterable[TracedItem]) -> None:
    """Extends the sequence with a list of items."""
    for item in items:
      self.append(item)

  @property
  def usage_summary(self) -> lf.UsageSummary:
    """Returns the usage summary of the action."""
    return self._usage_summary

  def execution_summary(self) -> dict[str, Any]:
    """Execution summary string."""
    return pg.Dict(
        subtree=dict(
            num_actions=len(self.all_actions),
            num_action_failures=len([
                a for a in self.all_actions if a.has_error
            ]),
            num_queries=len(self.all_queries),
            num_oop_failures=len([
                q for q in self.all_queries if q.has_oop_error
            ]),
            num_non_oop_failures=len([
                q for q in self.all_queries
                if q.has_error and not q.has_oop_error
            ]),
            total_query_time=sum(q.elapse for q in self.all_queries),
        ),
        current_level=dict(
            num_actions=len(self.actions),
            num_action_failures=len([
                a for a in self.actions if a.has_error
            ]),
            num_queries=len(self.queries),
            num_oop_failures=len([
                q for q in self.queries if q.has_oop_error
            ]),
            num_non_oop_failures=len([
                q for q in self.queries
                if q.has_error and not q.has_oop_error
            ]),
            execution_breakdown=[
                dict(
                    action=action.action.__class__.__name__,
                    usage=dict(
                        total_tokens=action.usage_summary.total.total_tokens,
                        estimated_cost=action.usage_summary.total.estimated_cost,
                    ),
                    execution_time=action.execution.elapse,
                )
                for action in self.actions
            ]
        )
    )

  #
  # HTML views.
  #

  def _html_tree_view_summary(
      self,
      *,
      name: str | None = None,
      extra_flags: dict[str, Any] | None = None,
      view: pg.views.html.HtmlTreeView, **kwargs
  ):
    return None

  def _execution_badge(self, interactive: bool = True):
    if not self.has_started:
      label = '(Not started)'
      tooltip = 'Execution not started.'
      css_class = 'not-started'
    elif not self.has_stopped:
      label = 'Starting'
      tooltip = 'Execution starting.'
      css_class = 'running'
    else:
      label = f'{int(self.elapse)} seconds'
      tooltip = pg.format(self.execution_summary(), verbose=False)
      css_class = 'finished'
    time_badge = pg.views.html.controls.Badge(
        label,
        tooltip=tooltip,
        css_classes=['execution-time', css_class],
        interactive=interactive,
    )
    if interactive:
      self._time_badge = time_badge
    return time_badge

  def _html_tree_view_content(
      self,
      *,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    del kwargs
    extra_flags = extra_flags or {}
    interactive = extra_flags.get('interactive', True)
    if interactive or self.items:
      self._tab_control = pg.views.html.controls.TabControl(
          [self._execution_item_tab(item) for item in self.items],
          tab_position='left'
      )
      return self._tab_control.to_html()
    return '(no tracked items)'

  def _execution_item_tab(self, item: TracedItem) -> pg.views.html.controls.Tab:
    if isinstance(item, ActionInvocation):
      css_class = 'action'
    elif isinstance(item, lf_structured.QueryInvocation):
      css_class = 'query'
    elif isinstance(item, lf.logging.LogEntry):
      css_class = f'log-{item.level}'
    elif isinstance(item, ExecutionTrace):
      css_class = 'phase'
    elif isinstance(item, ParallelExecutions):
      css_class = 'parallel'
    else:
      raise ValueError(f'Unsupported item type: {type(item)}')

    return pg.views.html.controls.Tab(
        label=self._execution_item_label(item),
        content=pg.view(item),
        css_classes=[css_class]
    )

  def _execution_item_label(
      self, item: TracedItem
  ) -> pg.views.html.controls.Label:
    if isinstance(item, ActionInvocation):
      return pg.views.html.controls.Label(
          item.action.__class__.__name__,
          tooltip=f'[{item.id}] Action invocation',
      )
    elif isinstance(item, lf_structured.QueryInvocation):
      schema_title = 'str'
      if item.schema:
        schema_title = lf_structured.annotation(item.schema.spec)
      return pg.views.html.controls.Label(
          schema_title,
          tooltip=f'[{item.id}] lf.Query invocation'
      )
    elif isinstance(item, lf.logging.LogEntry):
      return pg.views.html.controls.Label(
          item.level.title(),
          tooltip=item.message,
      )
    elif isinstance(item, ExecutionTrace):
      return pg.views.html.controls.Label(
          item.name or 'Phase',
          tooltip=f'[{item.id}] Execution group {item.name!r}'
      )
    elif isinstance(item, ParallelExecutions):
      return pg.views.html.controls.Label(
          item.name or 'Parallel',
          tooltip=f'[{item.id}] Parallel executions'
      )
    else:
      raise ValueError(f'Unsupported item type: {type(item)}')

  def _html_tree_view_css_styles(self) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .tab-button.action > ::before {
          content: "A";
          font-weight: bold;
          color: red;
          padding: 10px;
        }
        .tab-button.phase > ::before {
          content: "G";
          font-weight: bold;
          color: purple;
          padding: 10px;
        }
        .tab-button.parallel > ::before {
          content: "P";
          font-weight: bold;
          color: blue;
          padding: 10px;
        }
        .tab-button.query > ::before {
          content: "Q";
          font-weight: bold;
          color: orange;
          padding: 10px;
        }
        .tab-button.log-debug > ::before {
          content: "🔍";
          padding: 7px;
        }
        .tab-button.log-info > ::before {
          content: "ⓘ";
          color: blue;
          padding: 7px;
        }
        .tab-button.log-warning > ::before {
          content: "❗";
          padding: 7px;
        }
        .tab-button.log-error > ::before {
          content: "‼️";
          padding: 7px;
        }
        .tab-button.log-fatal > ::before {
          content: "❌";
          padding: 7px;
        }
        .details.execution-trace, .details.action-invocation {
          border: 1px solid #eee;
        }
        .execution-trace-title {
          display: inline-block;
        }
        .badge.execution-time {
          margin-left: 4px;
          border-radius: 0px;
        }
        .execution-time.starting {
          background-color: ghostwhite;
          font-weight: normal;
        }
        .execution-time.running {
          background-color: ghostwhite;
          font-weight: normal;
        }
        .execution-time.finished {
          background-color: aliceblue;
          font-weight: bold;
        }
       """
    ]


class ParallelExecutions(pg.Object, pg.views.html.HtmlTreeView.Extension):
  """A class for encapsulating parallel execution traces."""

  name: Annotated[
      str | None,
      'The name of the parallel execution.'
  ] = None

  branches: Annotated[
      list[ExecutionTrace],
      'The branches of the parallel execution.'
  ] = []

  def __getitem__(self, key: int) -> ExecutionTrace:
    return self.branches[key]

  def __len__(self) -> int:
    return len(self.branches)

  def _on_bound(self):
    super()._on_bound()
    self._tab_control = None
    self._lock = threading.Lock()

  def _on_parent_change(self, *args, **kwargs):
    super()._on_parent_change(*args, **kwargs)
    self.__dict__.pop('id', None)

  @functools.cached_property
  def id(self) -> str:
    parent = self.sym_parent
    if isinstance(parent, pg.List):
      container = parent.sym_parent
      if isinstance(container, ExecutionTrace):
        parallel_id = (
            self.name or f'p{container.indexof(self, ParallelExecutions) + 1}'
        )
        return f'{container.id}/{parallel_id}'
    return ''

  def add(self) -> ExecutionTrace:
    """Appends a branch to the parallel execution."""
    with self._lock, pg.notify_on_change(False):
      branch = ExecutionTrace(name=f'[{len(self)}]')
      self.branches.append(branch)
      if self._tab_control is not None:
        self._tab_control.append(self._branch_tab(branch))
      return branch

  #
  # HTML views.
  #

  def _html_tree_view_summary(
      self,
      *,
      name: str | None = None,
      extra_flags: dict[str, Any] | None = None,
      view: pg.views.html.HtmlTreeView, **kwargs
  ):
    return None

  def _html_tree_view_content(
      self,
      *,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    del kwargs
    extra_flags = extra_flags or {}
    interactive = extra_flags.get('interactive', True)
    if interactive or self.branches:
      self._tab_control = pg.views.html.controls.TabControl(
          [self._branch_tab(branch) for branch in self.branches],
          tab_position='left'
      )
      return self._tab_control.to_html()
    return '(no tracked parallel executions)'

  def _branch_tab(self, branch: ExecutionTrace) -> pg.views.html.controls.Tab:
    return pg.views.html.controls.Tab(
        label=pg.views.html.controls.Label(
            branch.name,
            tooltip=f'[{branch.id}] Branch {branch.name!r}'
        ),
        content=pg.view(branch),
    )


class ActionInvocation(pg.Object, pg.views.html.HtmlTreeView.Extension):
  """A class for capturing the invocation of an action."""
  action: Action

  result: Annotated[
      Any,
      'The result of the action.'
  ] = None

  metadata: Annotated[
      dict[str, Any],
      'The metadata returned by the action.'
  ] = {}

  error: Annotated[
      pg.ErrorInfo | None,
      'Error from the action if failed.'
  ] = None

  max_execution_time: Annotated[
      float | None,
      (
          'The maximum allowed execution time for the action. '
          'It is set when the action or any of its parent actions has been '
          'called with a `max_execution_time` argument. '
          'The value is the remaining time of the parent invocation or '
          'the `max_execution_time` for the current invocation, '
          'whichever is smaller. Since we are running action within the '
          'current thread, there is no guarantee that the action will be '
          'executed within this time limit, but we will try to stop the action '
          'as soon as the next operation on the session is called.'
      )
  ] = None

  execution: Annotated[
      ExecutionTrace,
      'The execution sequence of the action.'
  ] = ExecutionTrace()

  # Allow symbolic assignment without `rebind`.
  allow_symbolic_assignment = True

  def _on_bound(self):
    super()._on_bound()
    self._tab_control = None
    self.action._invocation = self    # pylint: disable=protected-access

  def _on_parent_change(self, *args, **kwargs):
    super()._on_parent_change(*args, **kwargs)
    self.__dict__.pop('id', None)

  @property
  def parent_action(self) -> Optional['ActionInvocation']:
    """Returns the parent action invocation."""
    return self.sym_ancestor(lambda x: isinstance(x, ActionInvocation))

  @property
  def max_remaining_execution_time(self) -> float | None:
    """Returns the remaining execution time for the action."""
    if self.max_execution_time is None:
      return None
    return max(0, self.max_execution_time - self.execution.elapse)

  @functools.cached_property
  def id(self) -> str:
    """Returns the id of the action invocation."""
    parent = self.sym_parent
    if isinstance(parent, Session):
      return f'{parent.id}:'
    elif isinstance(parent, pg.List):
      container = parent.sym_parent
      if isinstance(container, ExecutionTrace):
        action_id = f'a{container.indexof(self, ActionInvocation) + 1}'
        return f'{container.id}/{action_id}'
    return ''

  @property
  def has_error(self) -> bool:
    """Returns True if the action invocation has an error."""
    return self.error is not None

  @property
  def logs(self) -> list[lf.logging.LogEntry]:
    """Returns immediate child logs from execution sequence."""
    return self.execution.logs

  @property
  def actions(self) -> list['ActionInvocation']:
    """Returns immediate child action invocations."""
    return self.execution.actions

  @property
  def queries(self) -> list[lf_structured.QueryInvocation]:
    """Returns immediate queries made by the action."""
    return self.execution.queries

  @property
  def all_queries(self) -> list[lf_structured.QueryInvocation]:
    """Returns all queries made by the action and its child execution items."""
    return self.execution.all_queries

  @property
  def all_actions(self) -> list['ActionInvocation']:
    """Returns all actions made by the action and its child execution items."""
    return self.execution.all_actions

  @property
  def all_logs(self) -> list[lf.logging.LogEntry]:
    """Returns all logs made by the action and its child execution items."""
    return self.execution.all_logs

  @property
  def usage_summary(self) -> lf.UsageSummary:
    """Returns the usage summary of the action."""
    return self.execution.usage_summary

  @property
  def elapse(self) -> float:
    """Returns the elapsed time of the action."""
    return self.execution.elapse

  def start(self) -> None:
    """Starts the execution of the action."""
    self.execution.start()

  def end(
      self,
      result: Any,
      error: pg.ErrorInfo | None = None,
      metadata: dict[str, Any] | None = None,
  ) -> None:
    """Ends the execution of the action with result and metadata."""
    rebind_dict = dict(result=result, error=error)
    if metadata is not None:
      rebind_dict['metadata'] = metadata
    self.rebind(**rebind_dict, skip_notification=True, raise_on_no_change=False)
    self.execution.stop()
    if self._tab_control is not None:
      if self.metadata:
        self._tab_control.insert(
            1,
            pg.views.html.controls.Tab(
                'metadata',
                pg.view(
                    self.metadata,
                    collapse_level=None,
                    enable_summary_tooltip=False
                ),
                name='metadata',
            )
        )
      if self.has_error:
        self._tab_control.insert(
            1,
            pg.views.html.controls.Tab(
                'error',
                pg.view(
                    self.error,
                    collapse_level=None,
                    enable_summary_tooltip=False
                ),
                name='error',
            )
        )
      else:
        self._tab_control.insert(
            1,
            pg.views.html.controls.Tab(
                'result',
                pg.view(
                    self.result,
                    collapse_level=None,
                    enable_summary_tooltip=False
                ),
                name='result',
            ),
        )
      self._tab_control.select(['error', 'metadata', 'result'])

  #
  # HTML views.
  #

  def _html_tree_view_summary(
      self, *, view: pg.views.html.HtmlTreeView, **kwargs
  ):
    return None

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.html.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    extra_flags = extra_flags or {}
    interactive = extra_flags.get('interactive', True)
    if (isinstance(self.action, RootAction)
        and self.execution.has_stopped
        and len(self.execution) == 1):
      return view.content(self.execution.items[0], extra_flags=extra_flags)

    tabs = []
    if not isinstance(self.action, RootAction):
      tabs.append(
          pg.views.html.controls.Tab(
              'action',
              view.render(  # pylint: disable=g-long-ternary
                  self.action,
                  collapse_level=None,
                  root_path=self.action.sym_path,
                  enable_summary_tooltip=False,
              ),
              name='action',
          )
      )
    if self.execution.has_stopped:
      tabs.append(
          pg.views.html.controls.Tab(
              'result',
              view.render(
                  self.result,
                  collapse_level=None,
                  enable_summary_tooltip=False
              ),
              name='result'
          )
      )
      if self.metadata:
        tabs.append(
            pg.views.html.controls.Tab(
                'metadata',
                view.render(
                    self.metadata,
                    collapse_level=None,
                    enable_summary_tooltip=False
                ),
                name='metadata'
            )
        )

    tabs.append(
        pg.views.html.controls.Tab(
            pg.Html.element(
                'span',
                [
                    'execution',
                    self.execution._execution_badge(interactive),  # pylint: disable=protected-access
                    (
                        self.usage_summary.to_html(  # pylint: disable=g-long-ternary
                            extra_flags=dict(as_badge=True)
                        )
                    ),
                ],
                css_classes=['execution-tab-title']
            ),
            view.render(self.execution, extra_flags=extra_flags),
            name='execution',
        )
    )
    tab_control = pg.views.html.controls.TabControl(tabs)
    # Select the tab following a priority: metadata, result, action, execution.
    tab_control.select(['metadata', 'result', 'action', 'execution'])
    if interactive:
      self._tab_control = tab_control
    return tab_control

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .execution-tab-title {
          text-align: left;
        }
        .execution-tab-title .usage-summary.label {
          border-radius: 0px;
          font-weight: normal;
          color: #AAA;
        }
        """
    ]


class RootAction(Action):
  """A placeholder action for the root of the action tree."""

  def call(self, session: 'Session', **kwargs) -> Any:
    raise NotImplementedError('Shall not be called.')


class Session(pg.Object, pg.views.html.HtmlTreeView.Extension):
  """Session for performing an agentic task."""

  root: Annotated[
      ActionInvocation,
      'The root action invocation of the session.'
  ] = ActionInvocation(RootAction())

  id: Annotated[
      str | None,
      'An optional identifier for the sessin, which will be used for logging.'
  ] = None

  verbose: Annotated[
      bool,
      (
          'If True, the session will be logged with verbose action and query '
          'activities.'
      )
  ] = False

  #
  # Shortcut methods for accessing the root action invocation.
  #

  @property
  def all_queries(self) -> list[lf_structured.QueryInvocation]:
    """Returns all queries made by the session."""
    return self.root.all_queries

  @property
  def all_actions(self) -> list[ActionInvocation]:
    """Returns all actions made by the session."""
    return self.root.all_actions

  @property
  def all_logs(self) -> list[lf.logging.LogEntry]:
    """Returns all logs made by the session."""
    return self.root.all_logs

  @property
  def usage_summary(self) -> lf.UsageSummary:
    """Returns the usage summary of the session."""
    return self.root.usage_summary

  @property
  def has_started(self) -> bool:
    """Returns True if the session has started."""
    return self.root.execution.has_started

  @property
  def has_stopped(self) -> bool:
    """Returns True if the session has stopped."""
    return self.root.execution.has_stopped

  @property
  def has_error(self) -> bool:
    """Returns True if the session has an error."""
    return self.root.has_error

  @property
  def final_result(self) -> Any:
    """Returns the final result of the session."""
    return self.root.result

  @property
  def final_error(self) -> pg.ErrorInfo | None:
    """Returns the error of the session."""
    return self.root.error

  @property
  def elapse(self) -> float:
    """Returns the elapsed time of the session."""
    return self.root.elapse

  # NOTE(daiyip): Action execution may involve multi-threading, hence current
  # action and execution are thread-local.

  @property
  def _current_action(self) -> ActionInvocation:
    """Returns the current invocation."""
    return getattr(self._tls, '__current_action__')

  @_current_action.setter
  def _current_action(self, value: ActionInvocation):
    setattr(self._tls, '__current_action__', value)

  @property
  def _current_execution(self) -> ExecutionTrace:
    """Returns the current execution."""
    return getattr(self._tls, '__current_execution__')

  @_current_execution.setter
  def _current_execution(self, value: ExecutionTrace):
    setattr(self._tls, '__current_execution__', value)

  def _on_bound(self):
    super()._on_bound()
    self._tls = threading.local()
    self._current_action = self.root
    self._current_execution = self.root.execution
    if self.id is None:
      self.rebind(
          id=f'agent@{uuid.uuid4().hex[-7:]}',
          skip_notification=True
      )

  def start(self) -> None:
    """Starts the session."""
    self.root.execution.start()

  def end(
      self,
      result: Any,
      error: pg.ErrorInfo | None = None,
      metadata: dict[str, Any] | None = None,
  ) -> None:
    """Ends the session."""
    if error is not None:
      self.error(
          f'Trajectory failed in {self.elapse:.2f} seconds.',
          error=error,
          metadata=metadata,
          keep=True,
      )
    elif self.verbose:
      self.info(
          f'Trajectory succeeded in {self.elapse:.2f} seconds.',
          result=result,
          metadata=metadata,
          keep=False,
      )
    self.root.end(result, error, metadata)

  def check_execution_time(self) -> None:
    """Checks the execution time of the current action."""
    action = self._current_action
    if action is not None and 0 == action.max_remaining_execution_time:
      # We raise error on the top-most action which has a time limit.
      current_action = action
      parent_action = current_action.parent_action
      while parent_action is not None and parent_action.max_execution_time:
        current_action = parent_action
        parent_action = current_action.parent_action
      raise ActionTimeoutError(
          f'Action {current_action.id} '
          f'({current_action.action.__class__.__name__}) has exceeded its '
          f'maximum execution time of {current_action.max_execution_time} '
          'seconds.'
      )

  def __enter__(self):
    """Enters the session."""
    self.start()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    """Exits the session."""
    # We allow users to explicitly end the session with specified result
    # and metadata.
    if self.root.execution.has_stopped:
      return

    if exc_val is not None:
      result, metadata = None, None
      error = pg.ErrorInfo.from_exception(exc_val)
    else:
      actions = self.root.actions
      if actions:
        result = actions[-1].result
        error = actions[-1].error
        metadata = actions[-1].metadata
      else:
        result, error, metadata = None, None, None
    self.end(result, error, metadata)

  #
  # Context-manager for information tracking.
  #

  @contextlib.contextmanager
  def track_action(
      self,
      action: Action,
      max_execution_time: float | None = None
  ) -> Iterator[ActionInvocation]:
    """Track the execution of an action."""
    if not self.root.execution.has_started:
      raise ValueError(
          'Please call `Session.start() / Session.end()` explicitly, '
          'or use `with Session(...) as session: ...` context manager to '
          'signal the start and end of the session.'
      )

    # Early terminate the action if the execution time is exceeded.
    self.check_execution_time()

    invocation = ActionInvocation(
        pg.maybe_ref(action),
        max_execution_time=self._child_max_execution_time(max_execution_time)
    )
    parent_action = self._current_action
    parent_execution = self._current_execution
    parent_execution.append(invocation)

    try:
      self._current_action = invocation
      self._current_execution = invocation.execution
      # Start the execution of the current action.
      self._current_action.start()
      if self.verbose:
        self.info(
            'Action execution started.',
            action=invocation.action,
            keep=False,
        )
      yield invocation
    finally:
      if invocation.has_error:
        self.warning(
            (
                f'Action execution failed in '
                f'{invocation.execution.elapse:.2f} seconds.'
            ),
            action=invocation.action,
            error=invocation.error,
            keep=True,
        )
      elif self.verbose:
        self.info(
            (
                f'Action execution succeeded in '
                f'{invocation.execution.elapse:.2f} seconds.'
            ),
            action=invocation.action,
            result=invocation.result,
            keep=False,
        )
      self._current_execution = parent_execution
      self._current_action = parent_action

  @contextlib.contextmanager
  def track_phase(self, name: str | None) -> Iterator[ExecutionTrace]:
    """Context manager for starting a new execution phase (group)."""
    parent_execution = self._current_execution
    if name is None:
      phase = parent_execution
    else:
      phase = ExecutionTrace(name=name)
      phase.start()
      parent_execution.append(phase)

    try:
      self._current_execution = phase
      yield phase
    finally:
      if phase is not parent_execution:
        phase.stop()
        self._current_execution = parent_execution

  @contextlib.contextmanager
  def track_queries(
      self,
      phase: str | None = None
  ) -> Iterator[list[lf_structured.QueryInvocation]]:
    """Tracks `lf.query` made within the context.

    Args:
      phase: The name of a new phase to track the queries in. If not provided,
        the queries will be tracked in the parent phase.

    Yields:
      A list of `lf.QueryInvocation` objects, each for a single `lf.query`
      call.
    """
    def _query_start(invocation: lf_structured.QueryInvocation):
      # Early terminate the action if the execution time is exceeded.
      self.check_execution_time()

      execution = self._current_execution
      invocation.rebind(
          id=f'{execution.id}/q{len(execution.queries) + 1}',
          skip_notification=False, raise_on_no_change=False
      )
      execution.append(invocation)
      if self.verbose:
        self.info(
            'Querying LLM started.',
            lm=invocation.lm.model_id,
            output_type=(
                lf_structured.annotation(invocation.schema.spec)
                if invocation.schema is not None else None
            ),
            keep=False,
        )

    def _query_end(invocation: lf_structured.QueryInvocation):
      self._current_execution.merge_usage_summary(invocation.usage_summary)
      if invocation.has_error:
        self.warning(
            (
                f'Querying LLM failed in '
                f'{time.time() - invocation.start_time:.2f} seconds.'
            ),
            lm=invocation.lm.model_id,
            output_type=(
                lf_structured.annotation(invocation.schema.spec)
                if invocation.schema is not None else None
            ),
            error=invocation.error,
            keep=True,
        )
      elif self.verbose:
        self.info(
            (
                f'Querying LLM succeeded in '
                f'{time.time() - invocation.start_time:.2f} seconds.'
            ),
            lm=invocation.lm.model_id,
            output_type=(
                lf_structured.annotation(invocation.schema.spec)
                if invocation.schema is not None else None
            ),
            keep=False,
        )

    with self.track_phase(phase), lf_structured.track_queries(
        include_child_scopes=False,
        start_callabck=_query_start,
        end_callabck=_query_end,
    ) as queries:
      try:
        yield queries
      finally:
        pass

  #
  # Operations with activity tracking.
  #

  def add_metadata(self, **kwargs: Any) -> None:
    """Adds metadata to the current invocation."""
    with pg.notify_on_change(False):
      self._current_action.metadata.update(kwargs)

  def query(
      self,
      prompt: Union[str, lf.Template, Any],
      schema: Union[
          lf_structured.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
      ] = None,
      default: Any = lf.RAISE_IF_HAS_ERROR,
      *,
      lm: lf.LanguageModel,
      examples: list[lf_structured.MappingExample] | None = None,
      **kwargs
      ) -> Any:
    """Calls `lf.query` and associates it with the current invocation.

    The following code are equivalent:

      Code 1:
      ```
      session.query(...)
      ```

      Code 2:
      ```
      with session.track_queries() as queries:
        output = lf.query(...)
      ```
    The former is preferred when `lf.query` is directly called by the action.
    If `lf.query` is called by a function that does not have access to the
    session, the latter should be used.

    Args:
      prompt: The prompt to query.
      schema: The schema to use for the query.
      default: The default value to return if the query fails.
      lm: The language model to use for the query.
      examples: The examples to use for the query.
      **kwargs: Additional keyword arguments to pass to `lf.query`.

    Returns:
      The result of the query.
    """
    with self.track_queries():
      return lf_structured.query(
          prompt,
          schema=schema,
          default=default,
          lm=lm,
          examples=examples,
          **kwargs
      )

  def concurrent_map(
      self,
      func: Callable[[Any], Any],
      parallel_inputs: Iterable[Any],
      *,
      phase: str | None = None,
      max_workers: int = 32,
      timeout: int | None = None,
      silence_on_errors: Union[
          Type[BaseException], tuple[Type[BaseException], ...], None
      ] = Exception
  ) -> Iterator[Any]:
    """Starts and tracks parallel execution with `lf.concurrent_map`."""
    self.check_execution_time()
    parallel_inputs = list(parallel_inputs)
    parallel_execution = ParallelExecutions(name=phase)
    self._current_execution.append(parallel_execution)
    parent_action = self._current_action

    def _map_single(input_value):
      execution = parallel_execution.add()

      # This happens on a new thread. Therefore, we update the thread-local
      # states from the parent thread.
      self._current_execution = execution
      self._current_action = parent_action
      execution.start()
      try:
        with self.track_queries():
          return func(input_value)
      finally:
        execution.stop()

    for input_value, result, error in lf.concurrent_map(
        _map_single,
        parallel_inputs,
        max_workers=max_workers,
        timeout=self._child_max_execution_time(timeout),
        silence_on_errors=silence_on_errors
    ):
      yield input_value, result, error

  def _child_max_execution_time(
      self, max_execution_time: float | None
  ) -> float | None:
    """Returns the max execution time for the child action."""
    max_remaining = None
    if self._current_action is not None:
      max_remaining = self._current_action.max_remaining_execution_time
    if max_remaining is None:
      return max_execution_time
    if max_execution_time is None:
      return max_remaining
    return min(max_remaining, max_execution_time)

  def _log(
      self,
      level: lf.logging.LogLevel,
      message: str,
      keep: bool,
      *,
      for_action: Action | ActionInvocation | None = None,
      **kwargs
  ) -> None:
    """Logs a message to the session.

    Args:
      level: The logging level.
      message: The message to log.
      keep: Whether to keep the log entry in the execution trace.
      for_action: The action to log the message for. If not provided, the
        current action will be used.
      **kwargs: Additional keyword arguments to pass to `lf.logging.log` as
        metadata to show.
    """
    # Early terminate the action if the execution time is exceeded.
    self.check_execution_time()

    execution = self._current_execution
    if for_action is None:
      for_action = self._current_action
    elif isinstance(for_action, Action):
      for_action = for_action.invocation
      assert for_action is not None, (
          f'Action must be called before it can be logged: {for_action}'
      )

    log_entry = lf.logging.log(
        level,
        (
            f'[{for_action.id} ({for_action.action.__class__.__name__})]: '
            f'{message}'
        ),
        **kwargs
    )
    if keep:
      execution.append(log_entry)

  def debug(
      self,
      message: str,
      keep: bool = True,
      *,
      for_action: Action | ActionInvocation | None = None,
      **kwargs
  ) -> None:
    """Logs a debug message to the session."""
    self._log(
        'debug', message, keep=keep, for_action=for_action, **kwargs
    )

  def info(
      self,
      message: str,
      keep: bool = True,
      *,
      for_action: Action | ActionInvocation | None = None,
      **kwargs
  ) -> None:
    """Logs an info message to the session."""
    self._log(
        'info', message, keep=keep, for_action=for_action, **kwargs
    )

  def warning(
      self,
      message: str,
      keep: bool = True,
      *,
      for_action: Action | ActionInvocation | None = None,
      **kwargs
  ) -> None:
    """Logs a warning message to the session."""
    self._log(
        'warning', message, keep=keep, for_action=for_action, **kwargs
    )

  def error(
      self,
      message: str,
      keep: bool = True,
      *,
      for_action: Action | ActionInvocation | None = None,
      **kwargs
  ):
    """Logs an error message to the session."""
    self._log(
        'error', message, keep=keep, for_action=for_action, **kwargs
    )

  def fatal(
      self,
      message: str,
      keep: bool = True,
      *,
      for_action: Action | ActionInvocation | None = None,
      **kwargs
  ):
    """Logs a fatal message to the session."""
    self._log(
        'fatal', message, keep=keep, for_action=for_action, **kwargs
    )

  def as_message(self) -> lf.AIMessage:
    """Returns the session as a message."""
    return lf.AIMessage(
        'Agentic task session.',
        result=self.root
    )

  @property
  def current_action(self) -> ActionInvocation:
    """Returns the current invocation."""
    return self._current_action

  #
  # HTML views.
  #

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.html.HtmlTreeView,
      **kwargs
  ):
    return view.content(self.root, **kwargs)

  @classmethod
  def _html_tree_view_config(cls):
    config = super()._html_tree_view_config()
    config.update(
        enable_summary_tooltip=False,
        enable_key_tooltip=False,
    )
    return config


# Register the logging functions to skip the source of the logging functions.
pg.logging.register_frame_to_skip([
    Session._log,   # pylint: disable=protected-access
    Session.debug,
    Session.info,
    Session.warning,
    Session.error,
    Session.fatal
])



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import os
import tempfile
import unittest
from langfun.core.llms import fake
from langfun.core.structured import function_generation
import pyglove as pg


class FunctionGenerationTest(unittest.TestCase):

  def test_generate_function(self):
    function_gen_lm_response = inspect.cleandoc("""
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        """)
    unittest_lm_response = inspect.cleandoc("""
        ```python
          [
            UnitTest(
              input={
                'items': [1, 2, 3, 4, 5],
                'target': 3
              },
              expected_output=2
            ),
            UnitTest(
              input={
                'items': [1, 2, 3, 4, 5],
                'target': 6
              },
              expected_output=-1
            )
          ]
        ```
        """)

    lm = fake.StaticSequence([unittest_lm_response, function_gen_lm_response])

    @function_generation.function_gen(lm=lm, unittest='auto')
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)
    self.assertEqual(linear_search.source(), function_gen_lm_response)

  def test_generate_function_without_unittest(self):
    function_gen_lm_response = inspect.cleandoc("""
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        """)

    lm = fake.StaticSequence([function_gen_lm_response])

    @function_generation.function_gen(lm=lm)
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)
    self.assertEqual(linear_search.source(), function_gen_lm_response)

  def test_custom_unittest_examples(self):
    function_gen_lm_response = inspect.cleandoc("""
        ```python
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        ```
        """)

    lm = fake.StaticSequence([function_gen_lm_response])

    custom_unittest = [(([1, 2, 3, 4, 5], 3), 2)]

    @function_generation.function_gen(lm=lm, unittest=custom_unittest)
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)

  def test_custom_unittest_fn(self):
    function_gen_lm_response = inspect.cleandoc("""
        ```python
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        ```
        """)

    lm = fake.StaticSequence([function_gen_lm_response])

    def _unittest_fn(func):
      assert func([1, 2, 3, 4, 5], 3) == 2
      assert func([1, 2, 3, 4, 5], 6) == -1

    custom_unittest = _unittest_fn

    @function_generation.function_gen(lm=lm, unittest=custom_unittest)
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)

  def test_load_function_from_cache_file(self):
    lm = fake.StaticSequence([])

    def _unittest_fn(func):
      assert func([1, 2, 3, 4, 5], 3) == 2
      assert func([1, 2, 3, 4, 5], 6) == -1

    cache_file_dir = tempfile.gettempdir()
    cache_file = os.path.join(cache_file_dir, 'cache_file.json')

    cache_key = """@function_generation.function_gen(
        lm=lm,
        unittest=_unittest_fn,
        cache_filename=cache_file,
    )
    def linear_search(items, target):  # pylint: disable=unused-argument
      \"\"\"Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      \"\"\""""
    cache_value = """
        ```python
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        ```
    """
    cache = pg.Dict()
    cache[cache_key] = cache_value
    cache.save(cache_file)

    @function_generation.function_gen(
        lm=lm,
        unittest=_unittest_fn,
        cache_filename=cache_file,
    )
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)
    self.assertEqual(linear_search(['a', 'b', 'c'], 'd'), -1)

  def test_empty_cache_file(self):
    function_gen_lm_response = inspect.cleandoc("""
        ```python
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        ```
        """)

    lm = fake.StaticSequence([function_gen_lm_response])

    def _unittest_fn(func):
      assert func([1, 2, 3, 4, 5], 3) == 2
      assert func([1, 2, 3, 4, 5], 6) == -1

    cache_file_dir = tempfile.gettempdir()
    cache_file = os.path.join(cache_file_dir, 'cache_file.json')

    @function_generation.function_gen(
        lm=lm,
        unittest=_unittest_fn,
        cache_filename=cache_file,
    )
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)

  def test_context_passthrough(self):

    class Number(pg.Object):
      value: int

    function_gen_lm_response = inspect.cleandoc("""
        ```python
        def add(a: Number, b: Number) -> Number:
            \"\"\"Adds two numbers together.\"\"\"
            return Number(a.value + b.value)
        ```
        """)

    lm = fake.StaticSequence(
        [function_gen_lm_response]
    )

    def _unittest_fn(func):
      assert func(Number(1), Number(2)) == Number(3)

    custom_unittest = _unittest_fn

    @function_generation.function_gen(
        lm=lm, unittest=custom_unittest, num_retries=1
    )
    def add(a: Number, b: Number) -> Number:  # pylint: disable=unused-argument
      """Adds two numbers together."""

    self.assertEqual(add(Number(2), Number(3)), Number(5))

  def test_siganture_check(self):
    incorrect_signature_lm_response = inspect.cleandoc("""
        ```python
        def dummy():
          pass
        ```
        """)
    function_gen_lm_response = inspect.cleandoc("""
        ```python
        def linear_search(items, target):
            \"\"\"
            Performs a linear search on a list to find a target value.

            Args:
                items (list): The list to search within.
                target: The value to search for.

            Returns:
                int: The index of the target value if found, otherwise -1.
            \"\"\"
            for i, item in enumerate(items):
                if item == target:
                    return i
            return -1
        ```
        """)

    lm = fake.StaticSequence(
        [incorrect_signature_lm_response, function_gen_lm_response]
    )

    def _unittest_fn(func):
      assert func([1, 2, 3, 4, 5], 3) == 2
      assert func([1, 2, 3, 4, 5], 6) == -1

    custom_unittest = _unittest_fn

    @function_generation.function_gen(
        lm=lm, unittest=custom_unittest, num_retries=2
    )
    def linear_search(items, target):  # pylint: disable=unused-argument
      """Performs a linear search on a list to find a target value.

      Args:
          items (list): The list to search within.
          target: The value to search for.

      Returns:
          int: The index of the target value if found, otherwise -1.
      """

    self.assertEqual(linear_search(['a', 'b', 'c'], 'c'), 2)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Experiment patching for Langfun evaluations."""

import inspect
from typing import Union
import langfun.core as lf
from langfun.core import llms as lf_llms
from langfun.core.eval import base
import pyglove as pg


#
# Program-based patchers.
#


def patch_member(cls, key, value, parent_key: str | None = None):
  """Patches a member of a class."""

  def _rebind_fn(k, v, p):
    if (
        isinstance(p, cls)
        and k.key == key
        and (parent_key is None or (p and p.sym_path.key == parent_key))
    ):
      if inspect.isfunction(value):
        return value(k, v, p)
      return value
    return v

  return _rebind_fn


def patch_lm(lm: Union[lf.LanguageModel, pg.hyper.OneOf]):  # pylint: disable=redefined-outer-name
  """Patches the LLM of evaluations."""
  return patch_member(base.Evaluable, "lm", lm)


def patch_parsing_lm(lm: Union[lf.LanguageModel, pg.hyper.OneOf]):  # pylint: disable=redefined-outer-name
  """Patches the parsing LLM of evaluations."""
  return patch_member(base.Evaluable, "parsing_lm", lm)


def patch_schema_fn(schema_fn: Union[pg.Functor, pg.hyper.OneOf]):
  """Patches the schema_fn of evaluations."""
  return patch_member(base.Evaluable, "schema_fn", schema_fn)


def patch_prompt(prompt: Union[str, lf.Template, pg.hyper.OneOf]):
  """Patches the prompt of evaluations."""
  return patch_member(base.Evaluable, "prompt", prompt)


def patch_inputs(inputs: Union[pg.Functor, pg.hyper.OneOf]):
  """Patches the inputs used in evaluations."""
  return patch_member(base.Evaluable, "inputs", inputs)


def patch_additional_args(**kwargs):
  """Patches additional_args."""

  def value_fn(k, unused_v, p):
    # We infer the symbolic value for the old args, as it might be a
    # contextual attribute referring to its containing object.
    old_args = p.sym_inferred(k.key)
    if old_args:
      old_args = dict(old_args)
      old_args.update(kwargs)
      return old_args
    return kwargs

  return patch_member(base.Evaluable, "additional_args", value_fn)


#
# String-based patching.
#

_NAMED_MODELS = {
    # GPT models.
    "gpt35turbo": lf_llms.Gpt35Turbo,
    "gpt35turbo16k": lf_llms.Gpt35Turbo16K,
    "gpt4": lf_llms.Gpt4,
    "gpt4turbo": lf_llms.Gpt4Turbo,
    # Anthropic models.
    "haiku": lf_llms.Claude3Haiku,
    "claude3haiku": lf_llms.Claude3Haiku,
    "opus": lf_llms.Claude3Opus,
    "claude3opus": lf_llms.Claude3Opus,
    "sonnet": lf_llms.Claude3Sonnet,
    "claude3sonnet": lf_llms.Claude3Opus,
}


def model_by_name(name: str) -> lf.LanguageModel:
  """Gets model by name."""
  name = name.strip().lower()
  if name in _NAMED_MODELS:
    return _NAMED_MODELS[name]()
  raise ValueError(f"Unknown model name: {name}")


@pg.patcher(auto_typing=True)
def lm(unused_eval, models: list[str]):
  """Patch the LM used for benchmarking."""
  return patch_lm(pg.oneof([model_by_name(name) for name in models]))


@pg.patcher(auto_typing=True)
def temperature(unused_eval, value: float):
  """Patch the temperature used for benchmarking."""
  return patch_member(lf.LMSamplingOptions, "temperature", value)


@pg.patcher(auto_typing=True)
def max_tokens(unused_eval, value: int | None):
  """Patch the temperature used for benchmarking."""
  return patch_member(lf.LMSamplingOptions, "max_tokens", value)



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base for OpenAI compatible models (including OpenAI)."""

from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import openai as openai_conversion  # pylint: disable=unused-import
from langfun.core.llms import rest
import pyglove as pg


@lf.use_init_args(['api_endpoint', 'model'])
class OpenAICompatible(rest.REST):
  """Base for OpenAI compatible models."""

  model: Annotated[
      str, 'The name of the model to use.',
  ] = ''

  @property
  def headers(self) -> dict[str, Any]:
    return {
        'Content-Type': 'application/json'
    }

  def _request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # Reference:
    # https://platform.openai.com/docs/api-reference/completions/create
    # NOTE(daiyip): options.top_k is not applicable.
    args = dict(
        n=options.n,
        top_logprobs=options.top_logprobs,
    )
    if self.model:
      args['model'] = self.model
    if options.logprobs:
      args['logprobs'] = options.logprobs
    if options.temperature is not None:
      args['temperature'] = options.temperature
    if options.max_tokens is not None:
      args['max_completion_tokens'] = options.max_tokens
    if options.top_p is not None:
      args['top_p'] = options.top_p
    if options.stop:
      args['stop'] = options.stop
    if options.random_seed is not None:
      args['seed'] = options.random_seed
    if options.reasoning_effort is not None:
      args['reasoning_effort'] = options.reasoning_effort
    return args

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request_args = self._request_args(sampling_options)

    # Users could use `metadata_json_schema` to pass additional
    # request arguments.
    json_schema = prompt.metadata.get('json_schema')
    if json_schema is not None:
      if not isinstance(json_schema, dict):
        raise ValueError(
            f'`json_schema` must be a dict, got {json_schema!r}.'
        )
      if 'title' not in json_schema:
        raise ValueError(
            f'The root of `json_schema` must have a `title` field, '
            f'got {json_schema!r}.'
        )
      request_args.update(
          response_format=dict(
              type='json_schema',
              json_schema=dict(
                  schema=json_schema,
                  name=json_schema['title'],
                  strict=True,
              )
          )
      )
      prompt.metadata.formatted_text = (
          prompt.text
          + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
          + pg.to_json_str(request_args['response_format'], json_indent=2)
      )

    # Prepare messages.
    messages = []

    def modality_check(chunk: str | lf.Modality) -> Any:
      if (isinstance(chunk, lf_modalities.Mime)
          and not self.supports_input(chunk.mime_type)):
        raise ValueError(
            f'Unsupported modality: {chunk!r}.'
        )
      return chunk

    # Users could use `metadata_system_message` to pass system message.
    system_message = prompt.metadata.get('system_message')
    if system_message:
      assert isinstance(system_message, lf.SystemMessage), type(system_message)
      messages.append(
          system_message.as_format(
              'openai', chunk_preprocessor=modality_check
          )
      )
    messages.append(
        prompt.as_format('openai', chunk_preprocessor=modality_check)
    )
    request = dict()
    request.update(request_args)
    request['messages'] = messages
    return request

  def _parse_choice(self, choice: dict[str, Any]) -> lf.LMSample:
    # Reference:
    # https://platform.openai.com/docs/api-reference/chat/object
    logprobs = None
    choice_logprobs = choice.get('logprobs')
    if choice_logprobs:
      logprobs = [
          (
              t['token'],
              t['logprob'],
              [(tt['token'], tt['logprob']) for tt in t['top_logprobs']],
          )
          for t in choice_logprobs['content']
      ]
    return lf.LMSample(
        lf.Message.from_value(choice['message'], format='openai'),
        score=0.0,
        logprobs=logprobs,
    )

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    """Returns a LMSamplingResult from a JSON response."""
    usage = json['usage']
    return lf.LMSamplingResult(
        samples=[self._parse_choice(choice) for choice in json['choices']],
        usage=lf.LMSamplingUsage(
            prompt_tokens=usage['prompt_tokens'],
            completion_tokens=usage['completion_tokens'],
            total_tokens=usage['total_tokens'],
            completion_tokens_details=usage.get(
                'completion_tokens_details', None
            ),
        ),
    )

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if (status_code == 413
        or (status_code == 400 and b'string_above_max_length' in content)):
      return lf.ContextLimitError(f'{status_code}: {content}')
    return super()._error(status_code, content)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun eval framework v2."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order
from langfun.core.eval.v2.experiment import Experiment
from langfun.core.eval.v2.experiment import Suite
from langfun.core.eval.v2.evaluation import Evaluation

from langfun.core.eval.v2.example import Example
from langfun.core.eval.v2.progress import Progress

from langfun.core.eval.v2.metric_values import MetricValue
from langfun.core.eval.v2.metric_values import Rate
from langfun.core.eval.v2.metric_values import Average
from langfun.core.eval.v2.metrics import Metric
from langfun.core.eval.v2 import metrics

from langfun.core.eval.v2.experiment import Run
from langfun.core.eval.v2.experiment import RunId
from langfun.core.eval.v2.experiment import Plugin
from langfun.core.eval.v2.experiment import Runner
from langfun.core.eval.v2 import runners

# Plugins
from langfun.core.eval.v2.checkpointing import BulkCheckpointer
from langfun.core.eval.v2.checkpointing import PerExampleCheckpointer
from langfun.core.eval.v2.reporting import HtmlReporter


# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python sandboxing."""

import abc
import os
import tempfile
from typing import Annotated

from langfun.core.coding.python import parsing
import pyglove as pg


class SandboxOutput(pg.Object):
  """Sandbox output."""

  stdout: Annotated[
      str,
      'The stdout of the sandbox execution.'
  ] = ''

  stderr: Annotated[
      str,
      'The stderr of the sandbox execution.'
  ] = ''

  output_files: Annotated[
      dict[str, bytes],
      'The output files of the sandbox execution.'
  ] = {}


class BaseSandbox(pg.Object):
  """Interface and partial implementation for Python sandbox."""

  def _on_bound(self):
    super()._on_bound()
    self._uploaded_files: dict[str, str] = {}

  def run(
      self,
      code: str,
      *,
      timeout: int | None = 30,
      **kwargs
  ) -> SandboxOutput:
    """Runs code in the sandbox. Raises pg.coding.CodeError if failed."""
    return self._run(self.normalize_code(code), timeout=timeout, **kwargs)

  def normalize_code(self, code: str) -> str:
    """Returns normalized code runnable in the sandbox."""
    for original_path, uploaded_path in self._uploaded_files.items():
      code = code.replace(original_path, uploaded_path)
    return parsing.clean(code)

  @abc.abstractmethod
  def _run(
      self,
      code: str,
      *,
      timeout: int | None = 30,
      **kwargs
  ) -> SandboxOutput:
    """Runs code in the sandbox. Raises pg.coding.CodeError if failed."""

  def upload(self, path: str) -> str:
    """Uploads a file to the sandbox. Returns the uploaded path."""
    uploaded_path = self._upload(path)
    self._uploaded_files[path] = uploaded_path
    return uploaded_path

  @abc.abstractmethod
  def _upload(self, path: str) -> str:
    """Uploads a file to the sandbox."""

  def setup(self) -> None:
    """Sets up the sandbox."""
    self._uploaded_files = {}
    self._setup()

  def _setup(self) -> None:
    """Sets up the sandbox."""

  def cleanup(self) -> None:
    """Cleans up the sandbox."""
    self._uploaded_files = {}
    self._cleanup()

  def _cleanup(self) -> None:
    """Cleans up the sandbox."""

  def __enter__(self):
    """Enters the sandbox."""
    self.setup()
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Exits the sandbox."""
    self.cleanup()


class MultiProcessingSandbox(BaseSandbox):
  """Sandbox using multiprocessing."""

  def _on_bound(self):
    super()._on_bound()
    self._working_dir = None

  @property
  def working_dir(self) -> str | None:
    """Returns the directory of the sandbox."""
    return self._working_dir

  def _setup(self) -> None:
    """Sets up the sandbox."""
    self._working_dir = tempfile.TemporaryDirectory()

  def _cleanup(self) -> None:
    """Cleans up the sandbox."""
    assert self._working_dir is not None
    self._working_dir.cleanup()

  def _run(
      self,
      code: str,
      *,
      timeout: int | None = 30,
      **kwargs
  ) -> SandboxOutput:
    """Runs code in the sandbox."""
    stdout = pg.coding.run(
        code, returns_stdout=True, sandbox=True, timeout=timeout
    )
    return SandboxOutput(stdout=stdout)

  def _upload(self, path: str) -> str:
    """Uploads a file to the sandbox."""
    if self._working_dir is None:
      raise ValueError('Sandbox is not set up.')

    # Upload the file to the sandbox directory.
    uploaded_path = os.path.join(self._working_dir.name, os.path.basename(path))
    with pg.io.open(path, 'r') as r:
      with pg.io.open(uploaded_path, 'w') as w:
        w.write(r.read())
    return uploaded_path



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common value types for evaluation metrics and metadata."""


import abc
from typing import Annotated, Any, Union
import pyglove as pg


class MetricValue(pg.Object):
  """Base class for metric values."""

  class DataPoint(pg.Object):
    """A data point for a metric value."""
    example_id: int
    value: float
    weight: float = 1.0

  # NOTE(daiyip): For evaluations, usually the number of examples is within 10K,
  # therefore it's beneficial to store all accumulated values with their example
  # IDs so we are able to track the individual examples that contributed to this
  # metric value. If this premise changes, we might consider using a more
  # efficient data structure.
  data_points: Annotated[
      list[DataPoint],
      'Accumulated computed values with example IDs and weights.'
  ] = []

  total: Annotated[
      int,
      'The total number of examples being evaluated. Including errors.'
  ] = 0

  def _on_bound(self):
    super()._on_bound()
    self._weighted_sum = sum(dp.value * dp.weight for dp in self.data_points)

  def reset(self) -> None:
    """Resets the value to its initial state."""
    self._sync_members(data_points=[], total=0)
    self._weighted_sum = 0.0

  def _sync_members(self, **kwargs) -> None:
    """Synchronizes the members of this object."""
    self.rebind(**kwargs, skip_notification=True, raise_on_no_change=False)

  def __float__(self) -> float:
    """Returns the float representation of this object."""
    if self.total == 0:
      return float('nan')
    return self.reduce()

  @abc.abstractmethod
  def reduce(self) -> float:
    """Reduces the accumulated values into a single value."""

  def increment_total(self, delta: int = 1) -> 'MetricValue':
    """Increments the total number of examples being evaluated."""
    self._sync_members(total=self.total + delta)
    return self

  def add(
      self,
      example_id: int,
      value: float,
      weight: float = 1.0,
      increment_total: bool = False,
  ) -> 'MetricValue':
    """Adds a value to the accumulated values."""
    self._weighted_sum += value * weight
    with pg.notify_on_change(False), pg.allow_writable_accessors(True):
      self.data_points.append(
          MetricValue.DataPoint(example_id, value, weight)
      )
      if increment_total:
        self.increment_total()
    return self

  def __gt__(self, other: Union['MetricValue', float]) -> bool:
    if isinstance(other, self.__class__):
      return float(self) > float(other)
    return float(self) > other

  def __lt__(self, other: Union['MetricValue', float]) -> bool:
    if isinstance(other, self.__class__):
      return float(self) < float(other)
    return float(self) < other

  def __eq__(self, other: Union['MetricValue', float]) -> bool:
    if isinstance(other, self.__class__):
      return super().__eq__(other)
    return float(self) == other

  def __nonzero__(self) -> bool:
    return float(self) != 0

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      *args,
      **kwargs
  ) -> str:
    if compact:
      return super().format(compact, *args, **kwargs)
    if self.total == 0:
      return 'n/a'
    if verbose:
      return (
          f'{self.scalar_repr()} ({len(self.data_points)}/{self.total})'
      )
    return self.scalar_repr()

  @abc.abstractmethod
  def scalar_repr(self) -> str:
    """Returns the format string for the value."""

  def _sym_nondefault(self) -> dict[str, Any]:
    """Overrides nondefault valuesso volatile values are not included."""
    return dict()


class Rate(MetricValue):
  """Representing a rate in range [0, 1]."""

  def reduce(self) -> float:
    return self._weighted_sum / self.total

  def scalar_repr(self):
    if self.total == 0:
      return 'n/a'
    return f'{self.reduce():.1%}'


class Average(MetricValue):
  """Average of a aggregated values."""

  def reduce(self) -> float:
    if not self.data_points:
      return float('nan')
    return self._weighted_sum / len(self.data_points)

  def scalar_repr(self):
    return f'{self.reduce():.3f}'



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenize the prompt for `lf.query`."""

from typing import Any, Type, Union

import langfun.core as lf
from langfun.core.structured import mapping
from langfun.core.structured import querying
from langfun.core.structured import schema as schema_lib
import pyglove as pg


def tokenize(
    prompt: Union[str, pg.Symbolic] | list[str | pg.Symbolic],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ] = None,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    **kwargs,
) -> list[tuple[str | bytes, int]]:
  """Tokenize the prompt for `lf.query`.

  Args:
    prompt: The prompt(s) based on which each completion will be scored.
    schema: The schema as the output type. If None, it will be inferred from
      the completions.
    lm: The language model used for scoring.
    examples: Fewshot exemplars used together with the prompt in getting the
      completions.
    protocol: The protocol for formulating the prompt based on objects.
    **kwargs: Keyword arguments that are referred by the prompt.

  Returns:
    A list of (text, token_id) tuples.
  """
  input_message = querying.query_prompt(
      prompt,
      schema,
      examples=examples,
      protocol=protocol,
      **kwargs,
  )
  if lm is None:
    lm_override = lf.get_contextual_override('lm')
    if lm_override is None:
      raise ValueError('`lm` must be specified or provided from `lf.context`.')
    lm = lm_override.value

  return lm.tokenize(input_message)


async def atokenize(
    prompt: Union[str, pg.Symbolic] | list[str | pg.Symbolic],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any], None
    ] = None,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    **kwargs,
) -> list[tuple[str | bytes, int]]:
  """Async version of `lf.tokenize`."""
  # TODO(daiyip): implement native async tokenization.
  return await lf.invoke_async(
      tokenize,
      prompt,
      schema,
      lm=lm,
      examples=examples,
      protocol=protocol,
      **kwargs,
  )
